import asyncio
from time import time
from .custom_exception import ExceededRateLimitError
from .storage import Storage


class GeneralRateLimiter:
    """
    Rate limiter for general purpose.

    Notes:
    ------
    - DB interaction is not asynchronous!!!
    """
    def __init__(
            self, storage: Storage,
            max_requests: int, time_window: int = 1,
            max_capacity: int = 32, cleanup_threshold: float = 10
    ):
        self.__storage = storage
        self.__max_requests = max_requests
        self.__time_window = time_window
        self.__capacity = max_capacity
        self.__cleanup_threshold = cleanup_threshold

    def check_limit(self, key: str) -> bool:
        """
        Checks if a `key` has exceeded the rate limit.

        If the `key`'s first request is being made or if the time window has passed since the first request,
        it resets the start time and number of requests. Otherwise, it increments the number of requests.

        Parameters
        ----------
        key : str
            The key to check the rate limit for.

        Returns
        -------
        bool
            True if the key has not exceeded the rate limit, False otherwise.
        """
        current_time = time()
        item = self.__storage.get(key)
        if item is None:
            item = {"start_time": current_time, "num_requests": 1}
            self.__storage.set(key, item)
            return True

        if current_time - item.get("start_time") > self.__time_window:
            item = {"start_time": current_time, "num_requests": 1}
            self.__storage.set(key, item)
            return True

        item["num_requests"] += 1

        self.__storage.set(key, item)
        return item["num_requests"] <= self.__max_requests

    def cleanup(self):
        keys = self.__storage.keys()
        if len(keys) <= self.__capacity:
            return None

        current_time = time()
        for key in keys:
            item = self.__storage.get(key)
            if current_time - item.get("start_time") > self.__cleanup_threshold:
                self.__storage.drop(key)
        return None

    def __call__(self, key: str) -> bool:
        return_value = self.check_limit(key)

        self.cleanup()

        return return_value

    @classmethod
    def general_rate_limiter(
            cls, storage: Storage,
            max_requests: int, time_window: int = 1,
            max_capacity: int = 32, cleanup_threshold: float = 0.1
    ):
        """
        Decorator to limit the number of requests to a function.

        Parameters
        ----------
        storage : Storage
            The storage to use to store the number of requests made.
        max_requests : int
            The maximum number of requests a client can make within the time window.
        time_window : int
            The time window in seconds in which the number of requests is limited, default is 1 second.
        max_capacity : int
            The maximum number of keys to store in the storage.
        cleanup_threshold : float
            The threshold to clean up the storage.

        Returns
        -------
        function : The decorated function.

        Raises
        ------
        ExceededRateLimitError : If the rate limit is exceeded.

        Notes:
        ------
        - The key is the name of the function by default.
        - The key can be passed as a keyword argument to the function if rate limiting is required for different keys.
        """

        def decorator(func):
            limiter = GeneralRateLimiter(storage, max_requests, time_window, max_capacity, cleanup_threshold)

            def wrapper(*args, **kwargs):
                key = kwargs.get("key")
                if not key:
                    key = f"{func.__name__}"
                if not limiter(key):
                    raise ExceededRateLimitError(
                        f"Rate limit exceeded. "
                        f"`{key}` was/had called more than {max_requests} requests per {time_window} seconds."
                    )
                return func(*args, **kwargs)

            return wrapper

        return decorator


class GeneralRateLimiter_with_Lock:
    """
    Rate limiter for general purpose.
    Core operations are guarded by asyncio.Lock().

    Notes:
    ------
    - DB interaction is not asynchronous!!!
    """
    def __init__(
            self, storage: Storage,
            max_requests: int, time_window: int = 1,
            max_capacity: int = 32, cleanup_threshold: float = 10
    ):
        self.__storage = storage
        self.__max_requests = max_requests
        self.__time_window = time_window
        self.__capacity = max_capacity
        self.__cleanup_threshold = cleanup_threshold
        self.__lock = asyncio.Lock()

    async def check_limit(self, key: str) -> bool:
        async with self.__lock:
            current_time = time()
            item = self.__storage.get(key)
            if item is None:
                item = {"start_time": current_time, "num_requests": 1}
                self.__storage.set(key, item)
                return True

            if current_time - item.get("start_time") > self.__time_window:
                item = {"start_time": current_time, "num_requests": 1}
                self.__storage.set(key, item)
                return True

            item["num_requests"] += 1

            self.__storage.set(key, item)
            return item["num_requests"] <= self.__max_requests

    async def cleanup(self):
        async with self.__lock:
            keys = self.__storage.keys()

            if len(keys) <= self.__capacity:
                return None

            current_time = time()
            for key in keys:
                item = self.__storage.get(key)
                if current_time - item.get("start_time") > self.__cleanup_threshold:
                    self.__storage.drop(key)
            return None

    async def __call__(self, key: str) -> bool:
        return_value = await self.check_limit(key)

        await self.cleanup()

        return return_value

    @classmethod
    def general_rate_limiter(
            cls, storage: Storage,
            max_requests: int, time_window: int = 1,
            max_capacity: int = 32, cleanup_threshold: float = 0.1
    ):
        """
        Decorator to limit the number of requests to a function.

        Parameters
        ----------
        storage : Storage
            The storage to use to store the number of requests made.
        max_requests : int
            The maximum number of requests a client can make within the time window.
        time_window : int
            The time window in seconds in which the number of requests is limited, default is 1 second.
        max_capacity : int
            The maximum number of keys to store in the storage.
        cleanup_threshold : float
            The threshold to clean up the storage.

        Returns
        -------
        function : The decorated function.

        Raises
        ------
        ExceededRateLimitError : If the rate limit is exceeded.

        Notes:
        ------
        - The key is the name of the function by default.
        - The key can be passed as a keyword argument to the function if rate limiting is required for different keys.
        """

        def decorator(func):
            limiter = GeneralRateLimiter_with_Lock(storage, max_requests, time_window, max_capacity, cleanup_threshold)

            async def wrapper(*args, **kwargs):
                key = kwargs.get("key")
                if not key:
                    key = f"{func.__name__}"
                if not await limiter(key):
                    raise ExceededRateLimitError(
                        f"Rate limit exceeded. "
                        f"`{key}` was/had called more than {max_requests} requests per {time_window} seconds."
                    )
                return await func(*args, **kwargs)

            return wrapper

        return decorator


if __name__ == "__main__":
    pass



from collections import defaultdict
from typing import Any
from .storage import Storage


class BasicStorage(Storage):
    """
    Basic storage class that stores key-value pairs in memory without any persistence.
    """

    def __init__(self):
        # print("BasicStorage init")
        f: float | None = None
        self.__memory = defaultdict(lambda: {"start_time": f, "num_requests": 0})

    def get(self, key: str):
        return self.__memory.get(key)

    def set(self, key: str, value: Any):
        self.__memory.update({key: value})

    def drop(self, key: str):
        self.__memory.pop(key, None)

    def clear(self):
        self.__memory.clear()

    def keys(self) -> list[str]:
        return list(self.__memory.keys())



import json
import os
import re
import sqlite3
from typing import Any
from .storage import Storage


class SQLite3_Storage(Storage):
    """
    SQLite3_Storage is a subclass of the Storage abstract base class.
    This class provides a concrete implementation of the Storage interface using SQLite3 as the storage system.
    It stores key-value pairs in a SQLite3 database.

    Attributes:
    db_path (str): The path to the SQLite3 database.
    table_name (str): The name of the table in the SQLite3 database.
    """

    def __init__(self, db_path: str, table_name: str = "storage", overwrite: bool = False):
        """
        Initializes a new instance of the SQLite3_Storage class.

        Args:
        db_path (str): The path to the SQLite3 database.
        table_name (str, optional): The name of the table in the SQLite3 database. Defaults to "storage".
        overwrite (bool, optional): If True, overwrites the existing database at db_path. Defaults to False.
        """
        self.db_path = db_path
        self.table_name = table_name
        self.init(db_path, table_name, overwrite)

    @classmethod
    def init(cls, db_path: str, table_name: str, overwrite: bool = False):
        """
        Initializes the SQLite3 database.

        Args:
        db_path (str): The path to the SQLite3 database.
        table_name (str): The name of the table in the SQLite3 database.
        overwrite (bool, optional): If True, overwrites the existing database at db_path. Defaults to False.
        """
        _ = [
            SQLite3_Storage.validate_db_path(db_path),
            SQLite3_Storage.validate_table_name(table_name)
        ]  # Exception will be raised if validation fails

        if overwrite:
            if os.path.exists(db_path):
                os.remove(db_path)
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute(f"CREATE TABLE IF NOT EXISTS {table_name} (key TEXT PRIMARY KEY, value TEXT)")
        conn.commit()

    @classmethod
    def validate_db_path(cls, db_path: str):
        if not isinstance(db_path, str):
            raise ValueError(f"Invalid database path: {db_path}")

        if not db_path:
            raise ValueError("Database path cannot be empty")

        if not os.path.basename(db_path):
            raise ValueError(f"Invalid database path: {db_path}")

    @classmethod
    def validate_table_name(cls, table_name: str):
        if not isinstance(table_name, str):
            raise ValueError(f"Invalid table name: {table_name}")

        if not table_name:
            raise ValueError("Table name cannot be empty")

        if re.search(r"^\w+$", table_name) is None:
            raise ValueError(f"Invalid table name: {table_name}")

    def get(self, key: str):
        """
        Retrieves the value associated with the given key from the SQLite3 database.

        Args:
        key (str): The key to retrieve the value for.

        Returns:
        Any: The value associated with the given key, or None if the key does not exist.
        """
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute(f"SELECT value FROM {self.table_name} WHERE key=?", (key,))
        result = cursor.fetchone()
        if result:
            return json.loads(result[0])
        return None

    def set(self, key: str, value: Any):
        """
        Sets the value for the given key in the SQLite3 database.

        Args:
        key (str): The key to set the value for.
        value (Any): The value to set.
        """
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute(f"INSERT OR REPLACE INTO {self.table_name} (key, value) VALUES (?, ?)",
                       (key, json.dumps(value)))
        conn.commit()

    def drop(self, key: str):
        """
        Deletes the key-value pair associated with the given key from the SQLite3 database.

        Args:
        key (str): The key to delete the value for.
        """
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute(f"DELETE FROM {self.table_name} WHERE key=?", (key,))
        conn.commit()

    def clear(self):
        """
        Deletes all key-value pairs from the SQLite3 database.
        """
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute(f"DELETE FROM {self.table_name}")
        conn.commit()

    def keys(self) -> list[str]:
        """
        Returns a list of all keys in the SQLite3 database.

        Returns:
        list[str]: A list of all keys in the SQLite3 database.
        """
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute(f"SELECT key FROM {self.table_name}")
        return [row[0] for row in cursor.fetchall()]



from setuptools import setup, find_packages

DESCRIPTION = 'Another Python package aim to offer "Rate Limiting" functionality for general use cases.'

# python3 setup.py sdist bdist_wheel
# twine upload --skip-existing dist/* --verbose

VERSION = '0.0.3.1'

setup(
    name='pygrl',
    version=VERSION,
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='jonah_whaler_2348',
    author_email='jk_saga@proton.me',
    license='MIT',
    install_requires=[],
    keywords=[
        'python', 'rate', 'limiter', 'rate limiter', 'rate limiting'
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
    ]
)



from abc import ABC, abstractmethod
from typing import Any


class Storage(ABC):
    """
    Abstract class for storage.

    Methods:
    --------
    get(key: str) -> Any
        Gets the value associated with the key.

    set(key: str, value: Any) -> None
        Sets the value associated with the key.
    """

    @abstractmethod
    def get(self, key: str) -> Any:
        """
        Gets the value associated with the key.

        Parameters
        ----------
        key : str
            The key to get the value for.

        Returns
        -------
        Any
            The value associated with the key.
        """
        pass

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """
        Sets the value associated with the key.

        Parameters
        ----------
        key : str
            The key to set the value for.
        value : Any
            The value to set.
        """
        pass

    @abstractmethod
    def drop(self, key: str):
        """
        Drops the value associated with the key.

        Parameters
        ----------
        key : str
            The key to drop the value for.
        """
        pass

    @abstractmethod
    def clear(self):
        """
        Clears all the values in the storage.
        """
        pass

    @abstractmethod
    def keys(self) -> list:
        """
        Returns all the keys in the storage.
        """
        pass
